#include <stdint.h>
#include "config.h"
#include "convert.h"
#include "dda.h"
#include "delay.h"
#include "dist.h"
#include "cmd.h"
#include "fan.h"
#include "heater.h"
#include "parse.h"
#include "state.h"
#include "uart.h"
#include "valve.h"

void cmd_dwell(int32_t ms);
void cmd_extruder(uint8_t t, int8_t s);
void cmd_home(int8_t x, int8_t y, int8_t z);
void cmd_move(int32_t *fxyze);
void cmd_report_position(void);
void cmd_report_temps(uint8_t t);
void cmd_set_advance(int32_t k);
void cmd_set_area(int32_t a, int32_t x, int32_t z);
void cmd_set_position(int32_t *fxyze);
void cmd_set_relative(int8_t v);
void cmd_set_temp(uint8_t t, int32_t s);
void cmd_set_units(int8_t u);
void cmd_stop(void);
void cmd_valve(uint8_t t, int8_t s);
void cmd_wait_temps(uint8_t t);

int32_t cmd[27];
#define CMD(A) cmd[A - 'A']
#define CMD_STAR cmd[26]

void cmd_process(void) {
    int32_t unknown_g = 0, unknown_m = 0, rs, FXYZES[6];
    uint8_t m;

    /* if we can sensibly check for transmission errors, do so */
    if (CMD('M') != 110 && s.n != UNSEEN && CMD('N') != s.n)
        rs = s.n;
    else if (CMD_STAR != UNSEEN && CMD_STAR != parse_chk)
        rs = CMD('N');
    else
        rs = UNSEEN;

    if (rs != UNSEEN) {
        uart_puts("rs ");
        uart_puti(rs);
        uart_putc('\n');
        return;
    } else if (CMD('N') != UNSEEN)
        s.n = CMD('N') + 1;

    /* if we have no command - assume the last G command */
    if (CMD('G') == UNSEEN && CMD('M') == UNSEEN && CMD('T') == UNSEEN)
        CMD('G') = s.g;

    /* keep track of tool in use */
    if (CMD('T') != UNSEEN)
        s.t = CMD('T');

    FXYZES[F] = CMD('F');
    FXYZES[X] = CMD('X');
    FXYZES[Y] = CMD('Y');
    FXYZES[Z] = CMD('Z');
    FXYZES[E] = CMD('E');
    FXYZES[S] = CMD('S');

    /* convert inches to mm */
    if (s.units == UNITS_INCHES) {
        FOREACH_FXYZE(i)
            if (FXYZES[i] != UNSEEN)
                FXYZES[i] = INCH_TO_MM(FXYZES[i]);
        FXYZES[S] = INCH2_TO_MM2(FXYZES[S]);
    }

    /* do command */
    if (CMD('G') != UNSEEN) {
        s.g = CMD('G');
        switch (s.g) {
        case 0:
        case 1: cmd_move(FXYZES); break; // move
        case 4: dda_drain(); cmd_dwell(CMD('P')); break; // dwell (p in ms)
        case 20: cmd_set_units(UNITS_INCHES); break; // set units to inches
        case 21: cmd_set_units(UNITS_MM); break; // set units to mm
        case 28: cmd_home(FXYZES[X] != UNSEEN,
                          FXYZES[Y] != UNSEEN,
                          FXYZES[Z] != UNSEEN); break; // home axis
        case 90: cmd_set_relative(0); break; // absolute positioning
        case 91: cmd_set_relative(1); break; // relative positioning
        case 92: cmd_set_position(FXYZES); break; // set position
        default: unknown_g = CMD('G'); break;
        }
    }

    m = CMD('M');
    if (CMD('M') != UNSEEN)
        switch (m) {
        case 105: // get temperatures - implemented below
        case 110: // set line number - implemented above
        case 114: // get position - implemented below
            break;

        case 0: // stop
            dda_drain();
        case 112: // emergency stop
            cmd_stop(); break;
        case 100: // set extruder area
            cmd_set_area(FXYZES[S], FXYZES[X], FXYZES[Z]); break;
        case 101: // extruder on
        case 103: // extruder off
            cmd_extruder(s.t, m == 101); break;
        case 104: // set extruder temperature
        case 109: // set bed temperature
        case 140: // set bed temperature
            dda_drain(); cmd_set_temp(m == 104 ? s.t : BED, CMD('S')); break;
        case 106: // fan on
        case 107: // fan off
            dda_drain(); fan(m == 106); break;
        case 6: // wait for temperatures
        case 116: // wait for temperatures
            dda_drain(); cmd_wait_temps(s.t); break;
        case 126: // valve open
        case 127: // valve close
            dda_drain(); cmd_valve(s.t, m == 126); cmd_dwell(CMD('P')); break;
        case 200: // set extruder advance
            dda_drain(); cmd_set_advance(CMD('S')); break;

        default: unknown_m = CMD('M'); break;
        }

    /* and send response */
    if (unknown_g == 0 && unknown_m == 0)
        uart_puts("ok");
    else {
        uart_puts("??");
        if (unknown_g != 0) {
            uart_puts(" G");
            uart_puti(unknown_g);
        }
        if (unknown_m != 0) {
            uart_puts(" M");
            uart_puti(unknown_m);
        }
    }
    switch (m) {
    case 105: dda_drain(); cmd_report_temps(s.t); break; // get temperatures
    case 114: cmd_report_position(); break; // get position
    }
    uart_putc('\n');
}

void cmd_dwell(int32_t p) {
    int16_t ms = p / PRECISION;
    if (p == UNSEEN)
        return;
    while (ms--)
        delay_ms(1);
}

void cmd_extruder(uint8_t t, int8_t i) {
    s.extruding = i;
}

void cmd_home(int8_t x, int8_t y, int8_t z) {
    int32_t dfxyze[5];

    if (!x && !y && !z)
        x = y = z = 1;

    dfxyze[F] = PRECISION*HOME_FEEDRATE_XY;
    dfxyze[X] = x ? -s.fxyze[X] : 0;
    dfxyze[Y] = y ? -s.fxyze[Y] : 0;
    dfxyze[Z] = 0;
    dfxyze[E] = 0;
    dda_q(dfxyze);

    dfxyze[F] = PRECISION*HOME_FEEDRATE_Z;
    dfxyze[X] = 0;
    dfxyze[Y] = 0;
    dfxyze[Z] = z ? -s.fxyze[Z] : 0;
    dda_q(dfxyze);
}

void cmd_move(int32_t *fxyze) {
    int32_t steps[5];

    if (fxyze[F] != UNSEEN)
        s.fxyze[F] = fxyze[F];

    steps[F] = s.fxyze[F];
    FOREACH_XYZE(i) {
        steps[i] = convert(fxyze[i], i);
        if (s.rel == 0)
            steps[i] -= s.fxyze[i];
        if (fxyze[i] == UNSEEN)
            steps[i] = 0;
    }
    if (fxyze[E] == UNSEEN && s.extruding) steps[E] = UNSEEN;

    dda_q(steps);
}

void cmd_report_position(void) {
    uart_report('C');
    FOREACH_XYZE(i) {
        int32_t x = convert(s.fxyze[i], INV + i);
        if (s.units == UNITS_INCHES)
            x = MM_TO_INCH(x);

        uart_report(i < E ? i + 'X' - X : 'E');
        uart_putd(x);
    }
}

void cmd_report_temps(uint8_t t) {
    uart_report('T');
    uart_puti(heater_get_temp(t));
    uart_report('B');
    uart_puti(heater_get_temp(BED));
}

void cmd_set_advance(int32_t k) {
    if (k != UNSEEN)
        s.extruder_advance = k * STEPS_PER_CUBIC_MM_E * PRECISION_EXTRUDER / PRECISION;
}

void cmd_set_area(int32_t a, int32_t x, int32_t z) {
    if (a != UNSEEN)
        s.extrusion_area = a;
    if (x != UNSEEN && z != UNSEEN)
        s.extrusion_area = x * z / PRECISION;
}

void cmd_set_position(int32_t *fxyze) {
    FOREACH_XYZE(i)
        if (fxyze[i] != UNSEEN)
            s.fxyze[i] = convert(fxyze[i], i);
}

void cmd_set_relative(int8_t v) {
    s.rel = v;
}

void cmd_set_temp(uint8_t t, int32_t s) {
    if (s != UNSEEN)
        heater_set_temp(t, s / PRECISION);
}

void cmd_set_units(int8_t u) {
    s.units = u;
}

void cmd_stop(void) {
    return;
    heater_set_temp(BED, 0);
    heater_set_temp(0, 0);
    for(;;)
        /* nothing */;
}

void cmd_valve(uint8_t t, int8_t i) {
    valve(t, i);
}

void cmd_wait_temps(uint8_t t) {
    while (!heater_ready())
        delay_ms(100);
}
