#include <stdint.h>
#include "config.h"
#include "convert.h"
#include "dda.h"
#include "delay.h"
#include "dist.h"
#include "divide.h"
#include "log.h"
#include "motor.h"
#include "state.h"
#include "uart.h"

#define STATE_IDLE 0
#define STATE_START 1
#define STATE_SPEEDCHANGE 2
#define STATE_DDA 3
#define STATE_WAIT 4
#define STATE_FINISHED 5

#define MIN_F (PRECISION * MIN_FEEDRATE)
#define MAX_F_XY (PRECISION * MAX_FEEDRATE_XY)
#define MAX_F_Z (PRECISION * MAX_FEEDRATE_Z)
#define MAX_A_XY ((uint32_t) (ACCELERATION_XY / (US_IN_S / PRECISION)))
#define MAX_A_Z ((uint32_t) (ACCELERATION_Z / (US_IN_S / PRECISION)))

#if DDA_BUFFER == 1 || DDA_BUFFER == 2 || DDA_BUFFER == 4 || DDA_BUFFER == 8 || DDA_BUFFER == 16 || DDA_BUFFER == 32 || DDA_BUFFER == 64 || DDA_BUFFER == 128
#else
#error DDA_BUFFER must be 1, 2, 4, 8, 16, 32, 64 or 128
#endif

#if DDA_SPEEDCHANGE_RATE == 1 || DDA_SPEEDCHANGE_RATE == 2 || DDA_SPEEDCHANGE_RATE == 4 || DDA_SPEEDCHANGE_RATE == 8 || DDA_SPEEDCHANGE_RATE == 16 || DDA_SPEEDCHANGE_RATE == 32 || DDA_SPEEDCHANGE_RATE == 64 || DDA_SPEEDCHANGE_RATE == 128 || DDA_SPEEDCHANGE_RATE == 256
#else
#error DDA_SPEEDCHANGE_RATE must be 1, 2, 4, 8, 16, 32, 64, 128 or 256
#endif

static struct dda {
    int32_t delta[5], d, f, a;
} q[DDA_BUFFER];
static uint8_t q_h, q_t, q_used = 0;


void dda_q_calculate_max_f_and_a(int32_t *fxyze, int32_t d) {
    int32_t d_xy, d_z, t_xy, t_z;

    d_xy = dist_steps_xy(fxyze);
    d_z = convert(fxyze[Z], INV_Z);
    if (d_z < 0)
        d_z = -d_z;

    t_xy = (d_xy << 8) / (MAX_F_XY >> 8);
    t_z = (d_z << 8) / (MAX_F_Z >> 8);
    if (t_xy < t_z)
        t_xy = t_z;
    if (t_xy < 1)
        t_xy = 1;
    q[q_t].f = ((d << 8) / t_xy) << 8;

    t_xy = d_xy / MAX_A_XY;
    t_z = d_z / MAX_A_Z;
    if (t_xy < t_z)
        t_xy = t_z;
    if (t_xy < 1)
        t_xy = 1;
    q[q_t].a = d / t_xy;
}

void dda_q(int32_t *fxyze) {
    int32_t d;

    while (q_used == DDA_BUFFER)
        dda_tick();

    if (fxyze[X] == 0 && fxyze[Y] == 0 && fxyze[Z] == 0) {
        if (fxyze[E] == 0 || fxyze[E] == UNSEEN)
            return;
        d = convert(fxyze[E], INV_E);
        if (d < 0)
            d = -d;
        q[q_t].f = MAX_F_XY;
        q[q_t].a = MAX_A_XY;
    } else {
        d = dist_steps(fxyze);
        if (fxyze[E] == UNSEEN)
            fxyze[E] = convert(d, E);
        dda_q_calculate_max_f_and_a(fxyze, d);
    }

    dda_tick();

    q[q_t].d = d;
    if (q[q_t].f > fxyze[F])
        q[q_t].f = fxyze[F];
    FOREACH_XYZE(i) {
        q[q_t].delta[i] = fxyze[i];
        s.fxyze[i] += fxyze[i];
    }

#ifdef DEBUG
    uart_puts("dda_q ");
    uart_puti(q_t);
    uart_puts(":");
    uart_puti(q_h);
    uart_puts("\t");
    FOREACH_XYZE(i) {
        uart_puts(" "); uart_puti(fxyze[i]);
    }
    uart_puts("\tf"); uart_putd(fxyze[F]);
    uart_puts("\td"); uart_putd(q[q_t].d);
    uart_puts("\tf"); uart_putd(q[q_t].f);
    uart_puts("\ta"); uart_puti(q[q_t].a*(US_IN_S/PRECISION));
    uart_puts("\n");
#endif

    q_t = (q_t + 1) & (DDA_BUFFER - 1);
    q_used++;
}

void dda_tick(void) {
    static int8_t state = STATE_IDLE, shift;
    static int32_t delta[5], error[5], max, df, max_f, a, advance, old_advance;
    static uint16_t dt;
    static int32_t steps_left = -1;
#ifdef DEBUG
    static int32_t count[5];
#endif
    static int32_t current_f, f_dash, step_decel;
    int32_t e;
#if DDA_SPEEDCHANGE_RATE > 1
    static int8_t tick;
#endif

    switch(state++) {
    case STATE_IDLE:
        if (q_used == 0) {
            state = STATE_IDLE;

            if (old_advance > 0) {
                motor_steps(E, -old_advance);
                old_advance = 0;
            }
        }
        return;

    case STATE_START:
        delay_tick_reset();

        max = 0;
        steps_left = 0;
        FOREACH_XYZE(i) {
            delta[i] = q[q_h].delta[i];
            motor_dir(i, delta[i] >= 0);
            if (delta[i] < 0)
                delta[i] = -delta[i];
            if (delta[i] != 0)
                motor_enable(i, 1);
            steps_left += delta[i];
            if (max < delta[i])
                max = delta[i];
        }

        step_decel = steps_left >> 1;
        current_f = 0;
        f_dash = 0;
        max_f = q[q_h].f;
        a = q[q_h].a;
        shift = approx_log(max_f) - DDA_SHIFT;
        if (shift < 0) shift = 0;
        delta[T] = q[q_h].d >> shift;
        if (max < delta[T])
            max = delta[T];

        e = -(max >> 1);
        FOREACH_TXYZE(i)
            error[i] = e;
        return;

    case STATE_SPEEDCHANGE:
        /* calculate amount of time until the next T step (in us)
         *
         * dt = ((US_IN_MIN / T_steps_per_mm) << shift) / feedrate
         * where
         *      T_steps_per_mm = PRECISION
         *      feedrate = current_f / PRECISION
         * need to use >>8 to avoid overflow
         */
        if (current_f < MIN_F)
            current_f = MIN_F;
#ifdef DDA_USE_APPROX_DIVIDE
        dt = approx_divide((US_IN_MIN>>8) << shift, current_f>>8);
#else
        dt = ((US_IN_MIN>>8) << shift) / (current_f>>8);
#endif

        /* calculate amount of change to feedrate over that step */
        df = a * dt;

#if DEBUG
        {
            static uint32_t old_f;
            if (current_f != old_f) {
                old_f = current_f;
                uart_puts("  f="); uart_putd(current_f);
                uart_puts(" dt="); uart_puti(dt);
                uart_puts(" df="); uart_putd(df);
                uart_puts("\n");
            }
        }
#endif
        return;

    case STATE_DDA:
        if (steps_left <= 0) {
            state = STATE_FINISHED;
            return;
        }

        state = STATE_DDA;
        FOREACH_TXYZE(a)
            error[a] += delta[a];

        FOREACH_XYZ(a)
            if (error[a] > 0) {
                error[a] -= max;
#ifdef DEBUG
                count[a]++;
#endif
                motor_step(a);
                steps_left--;
            }
        motor_step_done();

        e = 0;
        if (error[E] > 0) {
            error[E] -= max;
#ifdef DEBUG
            count[E]++;
#endif
            e++;
            steps_left--;
        }
        if (delta[E]) {
            /*
             * advance = STEPS_PER_CUBIC_MM_E * K * v^2
             * where
             *     v = f * area
             *     f = (current_f * 256 / PRECISION / 60)
             *     area = s.extrusion_area / PRECISION
             */

            /*
             * advance = (s.extruder_advance * rate^2) >> 16
             * rate = (current_f>>12 * s.extrusion_area) >> 16
             *      = (f*60*PRECISION>>12 * area*PRECISION) >> 16
             *      = (f*area * 60*PRECISION*PRECISION) >> 28
             * rate^2 = ((f*area)^2 * 60^2*PRECISION^4) >> 56
             *
             * therefore...
             * s.extruder_advance
             *     = ((STEPS_PER_CUBIC_MM * K) << 72)
             *       / (60^2 * PRECISION^4)
             *     = STEPS_PER_CUBIC_MM * K * PRECISION_ADVANCE
             */
            int32_t rate = ((current_f>>12) * s.extrusion_area) >> 16;
            advance = (s.extruder_advance * rate * rate) >> 16;

#if DEBUG
            if (advance != old_advance) {
                uart_puts("    f="); uart_putd(current_f);
                uart_puts(" area="); uart_putd(s.extrusion_area);
                uart_puts(" rate="); uart_puti(rate);
                uart_puts(" advance="); uart_puti(advance);
                uart_puts("\n");
            }
#endif
        } else {
            advance = 0;
        }
        motor_steps(E, e + advance - old_advance);
        old_advance = advance;

        if (error[T] > 0) {
            error[T] -= max;
            state = STATE_WAIT;
            delay_tick_in(dt);
        }
        return;

    case STATE_WAIT:
        if (!delay_ticked()) {
            state = STATE_WAIT;
            return;
        }

#if DDA_SPEEDCHANGE_RATE == 1
        state = STATE_SPEEDCHANGE;
#else
        if ((++tick & (DDA_SPEEDCHANGE_RATE - 1)) == 0)
            state = STATE_SPEEDCHANGE;
        else
            state = STATE_DDA;
#endif

        if (steps_left < step_decel) {
            if (current_f > 0) {
                if (current_f == max_f) {
                    current_f = f_dash;
                } else {
                    current_f -= df;
                    if (current_f < 0)
                        current_f = 0;
                }
            }
        } else if (current_f < max_f) {
            f_dash = current_f;
            current_f += df;
            if (current_f >= max_f) {
                current_f = max_f;
                step_decel = (step_decel << 1) - steps_left;
            }
        }
        return;

    case STATE_FINISHED:
#ifdef DEBUG
        uart_puts("dda_done ");
        uart_puti(q_h);
        uart_puts("\t");
        FOREACH_XYZE(i) {
            uart_puts(" ");
            uart_puti(count[i]);
            count[i] = 0;
        }
        uart_puts("\n");
#endif

        if (AUTO_DISABLE_X) motor_enable(X, 0);
        if (AUTO_DISABLE_Y) motor_enable(Y, 0);
        if (AUTO_DISABLE_Z) motor_enable(Z, 0);
        if (AUTO_DISABLE_E) motor_enable(E, 0);

        q_h = (q_h + 1) & (DDA_BUFFER - 1);
        q_used--;
        state = STATE_IDLE;
        return;
    }
}

void dda_drain(void) {
    while (q_used > 0)
        dda_tick();
}
