#include <libavr/bits.h>
#include <libavr/flash.h>
#include <libavr/interrupt.h>
#include <stdint.h>
#include "hw.h"
#include "../heater.h"
#include "heater_table.h"
#include "../state.h"

#define HEATER_HZ 20 // 10Hz per thermistor

volatile static struct {
    int16_t t_on, t_off, t_current;
} x[8];
volatile static int8_t adc_i, timer_i;

void heater_init(void) {
    /* Use timer 0 to fire an interrupt every 256 * 256 cycles, which
     * is ~244HZ for a 16MHz MCU.
     * The interrupt handle starts the ADC at ~HEATER_HZ.
     */
    TCCR0A = 1<<WGM01 | 0<<WGM00;
    TCCR0B = 0<<WGM02 | 1<<CS12 | 0<<CS11 | 0<<CS10;
    OCR0A = 0xFF;
    OCR0B = 0;
    TCNT0 = 0;
    TIMSK0 = 1<<OCIE0A;

    /* setup ADC for first conversion */
    adc_i = ADC_BED;
    ADMUX = 1<<REFS0 | ADC_BED;

    /* setup pins */
    CBI(DDR, ADC_BED); CBI(PORT, ADC_BED);
    CBI(DDR, ADC_EXTRUDER); CBI(PORT, ADC_EXTRUDER);
    SBI(DDR, BED); CBI(PORT, BED);
    SBI(DDR, EXTRUDER); CBI(PORT, EXTRUDER);
}

void heater_set_temp(uint8_t i, int16_t t) {
    i = i == BED ? ADC_BED : ADC_EXTRUDER;
    x[i].t_on = (t - 1) << 2;
    x[i].t_off = (t + 1) << 2;
}

int16_t heater_get_temp(uint8_t i) {
    i = i == BED ? ADC_BED : ADC_EXTRUDER;
    return x[i].t_current >> 2;
}

int8_t heater_ready(void) {
    return (x[ADC_BED].t_current >= x[ADC_BED].t_on)
        && (x[ADC_EXTRUDER].t_current >= x[ADC_EXTRUDER].t_on);
}

SIGNAL(TIMER0_COMPA) {
    int8_t a = timer_i;
    if (a > (F_CPU / 256 / 256 / HEATER_HZ)) {
        timer_i = 0;
        ADCSRA = 1<<ADEN | 1<<ADSC | 1<<ADIE | 0x7; /* 125kHz */;
    } else {
        timer_i = a + 1;
    }
}

SIGNAL(ADC) {
    int16_t t;

    t = flash_read16(&heater_table[ADC]) << 2;
    t += x[adc_i].t_current * 3;
    t /= 4;
    x[adc_i].t_current = t;

    switch(adc_i) {
    case ADC_BED:
        if (t < x[ADC_BED].t_on)
            SBI(PORT, BED);
        else if (t > x[ADC_BED].t_off)
            CBI(PORT, BED);
        break;

    case ADC_EXTRUDER:
        if (t < x[ADC_EXTRUDER].t_on)
            SBI(PORT, EXTRUDER);
        else if (t > x[ADC_EXTRUDER].t_off)
            CBI(PORT, EXTRUDER);
        break;
    }

    adc_i = (adc_i == ADC_BED) ? ADC_EXTRUDER : ADC_BED;
    ADMUX = 1<<REFS0 | adc_i;
}
